;fastshft.asm - Deliver a predetermined sequence of visual stimuli on
;alternating sides, with display updates synchronised to the refresh pulse.
;Put codes on the serial port for target stimuli and for joystick responses.
;Copyright (c) 2000 Matthew Belmonte
;based on fastshft.c by Matthew Belmonte
;
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;If you find this program useful, please send mail to Matthew Belmonte
;<mkb4@Cornell.edu>.  If you base a publication on data processed by this
;program, please notify Matthew Belmonte and include the following citation
;in your publication:
;
;	Matthew Belmonte, `A Software System for Analysis of
;	Steady-State Evoked Potentials', Association for Computing
;	Machinery SIGBIO Newsletter 17:1:9-14 (April 1997).

TITLE	  FASTSHFT		FAST SHIFT EXPERIMENT
CODESG	  SEGMENT PARA 'Code'
	  ASSUME CS:CODESG,DS:CODESG,SS:CODESG,ES:CODESG
	  ORG 100H		;START AT END OF PSP

;;;
;CONSTANTS
;;;

;PORT ADDRESSES
SYNCREG	  EQU 3DAH		;COLOUR SYNC REGISTER
GAMEPORT  EQU 201H		;GAME PORT
TIMEPORT  EQU  40H		;8253 PROGRAMMABLE INTERVAL TIMER
PPIPORT	  EQU  60H		;8255 PROGRAMMABLE PERIPHERAL INTERFACE
WAITPORT  EQU  80H		;WAIT PORT
;"Outputting any byte to port 0x80 should wait for almost exactly 1 microsecond
; independent of your processor type and speed" - Riku Saikkonen, Linux I/O
; port programming mini-HOWTO

;8253 PIT MASKS
TIM_CTR0  EQU 00H		;COUNTER 0 (IRQ0, VECTOR AT 1CH)
TIM_CTR1  EQU 40H		;COUNTER 1 (DRAM REFRESH)
TIM_CTR2  EQU 80H		;COUNTER 2 (SPEAKER)

TIM_LSB	  EQU 10H		;READ ONLY LSB OF COUNT
TIM_MSB	  EQU 20H		;READ ONLY MSB OF COUNT
TIM_WORD  EQU 30H		;READ LSB, THEN MSB

TIM_INTR  EQU 00H		;INTERRUPT
TIM_SHOT  EQU 02H		;ONE-SHOT
TIM_RATE  EQU 04H		;RATE GENERATOR
TIM_SQWV  EQU 06H		;SQUARE WAVE
TIM_SSTR  EQU 08H		;SOFTWARE-TRIGGERED STROBE
TIM_HSTR  EQU 0AH		;HARDWARE-TRIGGERED STROBE

TIM_BIN	  EQU 00H		;BINARY COUNT
TIM_BCD	  EQU 01H		;BCD COUNT

;8255 PPI MASKS
PPI_SENA  EQU 01H		;SPEAKER ENABLE (ENABLES 8253 TIMER 2 OUTPUT)
PPI_SDAT  EQU 02H		;SPEAKER DATA (ANDED WITH 8253 TIMER 2 OUTPUT)

;GAME PORT MASKS
HMASK	  EQU 01H		;RC TIMER ON PINS 3 AND 1
VMASK	  EQU 02H		;RC TIMER ON PINS 6 AND 1
B1MASK	  EQU 10H		;SWITCH ON PINS 2 AND 4
B2MASK	  EQU 20H		;SWITCH ON PINS 7 AND 4
B3MASK	  EQU 80H		;SWITCH ON PINS 14 AND 4

;PARALLEL PORT CONTROL CODES
LPT_DIR	  EQU 40H		;PS/2 DATA DIRECTION (0=OUTPUT, 1=INPUT)
LPT_IRQ	  EQU 20H		;IRQ ENABLE
LPT_STRB  EQU 01H		;STROBE MASK FOR CONTROL REGISTER (ACTIVE HIGH)

;PARALLEL PORT STATUS CODES
LPT_BUSY  EQU 80H		;BUSY MASK FOR STATUS REGISTER (ACTIVE LOW)
LPT_ACK	  EQU 40H		;ACK MASK FOR STATUS REGISTER (ACTIVE LOW)

;EVENT CODES
LBGCODE	  EQU 01H		;LEFT BACKGROUND
LTGTCODE  EQU 02H		;LEFT TARGET
RBGCODE	  EQU 03H		;RIGHT BACKGROUND
RTGTCODE  EQU 04H		;RIGHT TARGET
LBGNCODE  EQU 05H		;BEGIN TRIAL ATTENDING LEFT
RBGNCODE  EQU 06H		;BEGIN TRIAL ATTENDING RIGHT
ENDCODE	  EQU 07H		;END TRIAL
STRTCODE  EQU 84H		;START RECORDING
STOPCODE  EQU 88H		;STOP RECORDING

;COLOURS
TARGET	  EQU 04H		;TARGET SQUARE (USE 0CH FOR A LIGHTER RED)
NONTARG	  EQU 02H		;NONTARGET SQUARE (USE 0AH FOR A LIGHTER GREEN)
CAL	  EQU 0EH		;CALIBRATION TARGETS
FIX	  EQU 0FH		;FIXATION POINT
BGND	  EQU 00H		;BACKGROUND

;DIMENSIONS
NUMXPIX	  EQU 320		;HORIZONTAL SCREEN DIMENSION, MODE 13H
NUMYPIX	  EQU 200		;VERTICAL SCREEN DIMENSION, MODE 13H
SQ_HLEN	  EQU NUMYPIX/4		;HORIZONTAL STIMULUS DIMENSION (50)
SQ_VLEN	  EQU 86*SQ_HLEN/100	;CORRECT FOR PIXEL ASPECT RATIO (43)
SQ_LPOS	  EQU 0			;HORIZONTAL OFFSET OF LEFT EDGE OF LEFT TARGET
SQ_RPOS	  EQU NUMXPIX-1-SQ_HLEN	;HORIZONTAL OFFSET OF LEFT EDGE OF RIGHT TARGET

;TIMING
ONSET_DELAY  EQU 72*(12+30)	;12-SECOND STABILISATION + 30-SECOND BASELINE
OFFSET_DELAY EQU 72*30		;ANOTHER 30-SECOND BASELINE AT END

;FILE I/O
BUFSIZ	 EQU 100H		;FILE I/O BUFFER SIZE

;;;
;RUN THE EXPERIMENT.  READ CODES FROM THE FILE 'STIMULI.DAT' IN THE CURRENT
;DIRECTORY.  CODE 1 = LEFT NONTARGET, 2 = LEFT TARGET, 3 = RIGHT NONTARGET,
;4 = RIGHT TARGET, 5 = BEGIN LEFT TRIAL, 6 = BEGIN RIGHT TRIAL, 7 = END TRIAL.
;;;
MAIN:	  PUSHF			;SAVE INTERRUPT MASK
	  CLI			;MASK INTERRUPTS
	  CALL JOYSTK		;CALIBRATE JOYSTICK INPUT
	  MOV JOYOFF,BX
	  CALL PORTADDR		;RECORD I/O HARDWARE ADDRESSES
	  CALL SCONF		;CONFIGURE SERIAL PORT
	  CALL PCONF		;CONFIGURE PARALLEL PORT
	  CALL ACONF		;CONFIGURE AUDIO HARDWARE
	  MOV AH,0FH		;GET CURRENT VIDEO MODE
	  INT 10H
	  PUSH AX		;SAVE IT
	  MOV AX,0013H		;SET GRAPHICS MODE
	  INT 10H
	  CALL CLS		;CLEAR SCREEN
	  CALL CALIB		;PLOT CALIBRATION TARGETS
	  CALL FIXATE		;PLOT FIXATION POINT
	  SUB SP,BUFSIZ+8	;ALLOCATE BUFFER
	  MOV BX,SP		;POINT TO BUFFER
	  LEA DX,FILENAME	;POINT TO NAME OF DATA FILE
	  CALL FOPEN		;OPEN THE FILE
	  DEC SP
	  MOV DI,SP		;STACK A DUMMY ENDCODE FOR KBINTR
	  MOV BYTE PTR [DI],ENDCODE
	  MOV BP,SP		;BP WILL POINT PAST DATA ABOUT TO BE STACKED
READCODE: CALL SCANINT		;LOAD NEXT CODE INTO DL
	  JC DISPLAY		;(SKIP LOAD IF NO MORE CODES)
	  DEC SP		;MAKE ROOM ON STACK
	  MOV DI,SP		;POINT TO STACK
	  MOV [DI],DL		;PUSH CODE ONTO STACK
	  CMP AH,0FFH		;IF NOT AT EOF
	  JNE READCODE		;CONTINUE READING CODES.
DISPLAY:  MOV SI,BP		;POINT JUST IN FRONT OF FIRST CODE
DISPTEST: CMP SI,SP		;IF NO MORE CODES
	  JE DISPEND		;END DISPLAY
	  DEC SI		;OTHERWISE POINT TO NEXT CODE
	  MOV AL,[SI]		;AND LOAD IT.
	  CMP AL,RTGTCODE	;IF IT'S A STIMULUS CODE
	  JA DISP0
	  CALL PLOT		;DISPLAY IT (AND SEND CODE IF TARGET).
	  CALL KBINTR		;PROCESS ANY KEYBOARD INTERRUPT
	  JMP DISPTEST		;CONTINUE
DISP0:	  CMP AL,LBGNCODE	;IF IT'S A LEFT-ATTEND BEGIN
	  JNE DISP1
	  MOV CX,SQ_LPOS	;LEFT SQUARE,
	  MOV AX,0FF00H+LBGNCODE ;LEFT POSITION CODE AND EVENT CODE.
DISPBGN:  PUSH AX		;SAVE CODES FOR LATER USE
	  MOV DL,FIX		;USE FOREGROUND COLOUR
	  CALL SQUARE		;PLOT SQUARE AT POSITION [CX]
	  PUSH SI
	  MOV AH,00H		;WAIT FOR KEYBOARD INPUT
	  INT 16H
	  CMP AL,03H		;CHECK FOR CTRL-C ABORT:
	  JE DISPEND		;IF ABORTED, CLEAN UP STACK AND RETURN.
	  POP SI
	  CALL MRISYNC		;SOUND TONE TILL OPERATOR TRIGGERS SCANNER
	  MOV AL,STRTCODE	;START RECORDING
	  CALL SOUT
	  MOV DL,BGND		;BLANK
	  CALL SQUARE
	  MOV CX,ONSET_DELAY	;SHOULD BE AT LEAST 4 TO ALLOW RECORDER TO START
STIMON:	  CALL SYNC
	  LOOP STIMON
	  POP AX		;SEND BEGIN CODE.
	  CALL SOUT
	  JMP DISPTEST
DISP1:	  CMP AL,RBGNCODE	;IF IT'S A RIGHT-ATTEND BEGIN
	  JNE DISP2
	  MOV CX,SQ_RPOS	;RIGHT SQUARE,
	  MOV AX,0000H+RBGNCODE	;RIGHT POSITION AND EVENT CODE.
	  JMP DISPBGN
DISP2:	  CMP AL,ENDCODE	;IF IT'S AN END-OF-TRIAL
	  JNE DISP3
	  CALL SOUT		;SEND END CODE
	  MOV CX,OFFSET_DELAY	;SHOULD BE AT LEAST 4 TO ALLOW RECORDER TO STOP
STIMOFF:  CALL SYNC
	  LOOP STIMOFF
	  MOV AL,STOPCODE	;STOP RECORDING.
DISP3:	  CALL SOUT		;SEND CODE
	  JMP DISPTEST
DISPEND:  MOV SP,BP		;RESTORE STACK
	  ADD SP,BUFSIZ+9
	  POP AX		;RESTORE PREVIOUS VIDEO MODE
	  MOV AH,0
	  INT 10H
	  POPF			;RESTORE INTERRUPT MASK
	  MOV AX,4C00H		;SUCCESSFUL COMPLETION
	  INT 21H		;RETURN TO MSDOS
FILENAME  DB 'STIMULI.DAT',0

;;;
;PORTADDR - COPY PORT ADDRESSES FROM BIOS SEGMENT TO CURRENT DATA SEGMENT
;REFERENCE: Peter Abel, `IBM PC Assembly Language and Programming'
;(Prentice-Hall, 1998), page 479.
;MODIFIES AX,BX,DX, AND ALSO REGISTERS MODIFIED BY INT 21H
;;;
PORTADDR: MOV DX,DS		;SAVE CURRENT SEGMENT POINTER
	  MOV DS,0040H		;BIOS SEGMENT
	  MOV AX,[00H]		;GET ADDRESS OF BASE SERIAL PORT (COM1)
	  MOV BX,[10H]		;GET DEVICE COUNTS
	  AND BH,0CH		;MASK FOR HIGH-ORDER BITS OF SERIAL PORT COUNT
	  JE ONEPORT		;IF MORE THAN ONE SERIAL PORT IS INSTALLED,
	  MOV AX,[02H]		;GET ADDRESS OF FIRST EXPANSION PORT (COM2)
ONEPORT:  MOV BX,[08H]		;GET PARALLEL PORT ADDRESS
	  MOV DS,DX		;RESTORE SEGMENT POINTER
	  MOV SADDR,AX		;STORE SERIAL PORT ADDRESS
	  MOV PADDR,BX		;STORE PARALLEL PORT ADDRESS
	  RET

;;;
;SCONF - CONFIGURE SERIAL PORT FOR 19200bps 8N1
;MODIFIES AL,DX
;;;
SCONF:	  MOV DX,SADDR		;SERIAL PORT BASE ADDRESS
	  ADD DX,3		;POINT TO CONTROL REGISTER
	  MOV AL,83H		;ENABLE WRITES TO BAUD RATE REGISTER
	  OUT DX,AL
	  SUB DX,2
	  MOV AL,0		;BAUD RATE DIVISOR HIGH = 0
	  OUT DX,AL
	  DEC DX
	  MOV AL,06H		;BAUD RATE DIVISOR LOW = 6 (19200)
	  OUT DX,AL		;ALT. VALS: 0CH (9600), 03H (38400), 01H (56000)
	  ADD DX,3
	  MOV AL,03H		;DISABLE BAUD WRITES, SET 8N1
	  OUT DX,AL
	  RET

;;;
;PCONF - CONFIGURE PARALLEL PORT FOR OUTPUT
;MODIFIES AL,DX
;;;
PCONF:	  MOV DX,PADDR
	  MOV AL,0FFH		;SET ALL LINES HIGH TO POWER THE INTERFACE BOX
	  OUT DX,AL		;NOTE THAT POUT, IF USED, WILL INTERFERE W/ THIS
	  ADD DX,2		;POINT TO CONTROL REGISTER
	  IN AL,DX		;GET CURRENT SETTINGS
	  AND AL,0FFH-LPT_DIR-LPT_IRQ-LPT_STRB ;OUTPUT, NO IRQ, STROBE INACTIVE
	  OUT DX,AL		;MODIFY SETTINGS
	  RET

;;;
;ACONF - CONFIGURE AUDIO HARDWARE FOR 896Hz SQUARE-WAVE OUTPUT
;MODIFIES AX
;;;
ACONF:	  MOV AL,TIM_CTR2+TIM_WORD+TIM_SQWV+TIM_BIN ;SQUARE-WAVE OUTPUT
	  OUT TIMEPORT+3,AL
	  MOV AX,1331		;TIMER DIVISOR FOR 896Hz TONE
	  OUT TIMEPORT+2,AL	;WRITE LSB
	  MOV AL,AH
	  OUT TIMEPORT+2,AL	;WRITE MSB
	  RET

;;;
;SOUT - SEND AN EVENT CODE THROUGH THE SERIAL PORT
;ENTRY CONDITIONS: EVENT CODE IN AL
;MODIFIES AH,DX
;;;
SOUT:	  MOV DX,SADDR		;LOAD SERIAL PORT BASE ADDRESS
	  ADD DX,5		;POINT AT LINE STATUS REGISTER
	  MOV AH,AL
XMITWAIT: IN AL,DX		;GET STATUS BITS
	  AND AL,20H		;MASK FOR TRANSMITTER HOLDING REGISTER EMPTY
	  JE XMITWAIT		;WAIT TILL OK TO WRITE
	  MOV AL,AH
	  SUB DX,5		;POINT AT DATA REGISTER
	  OUT DX,AL		;SET OUTPUT
	  RET

;;;
;POUT - SEND AN EVENT CODE THROUGH THE PARALLEL PORT
;ENTRY CONDITIONS: EVENT CODE IN AL
;MODIFIES AX,DX
;;;
POUT:	  MOV DX,PADDR		;PARALLEL PORT BASE ADDRESS
	  INC DX		;POINT AT STATUS REGISTER
	  MOV AH,AL		;SAVE EVENT CODE
PBUSY:	  IN AL,DX		;READ STATUS REGISTER
	  TEST AL,LPT_BUSY	;TEST FOR DEVICE BUSY
	  OUT WAITPORT,AL	;DELAY 1us
	  JE PBUSY		;LOOP IF DEVICE WAS BUSY UP TO 1us AGO
PTIMEOUT: DEC DX		;POINT AT DATA REGISTER
	  MOV AL,AH		;RESTORE EVENT CODE
	  OUT DX,AL		;PLACE EVENT CODE ON DATA LINES
	  OUT WAITPORT,AL	;GIVE DATA TIME TO SETTLE
	  ADD DX,2		;POINT AT CONTROL REGISTER
	  IN AL,DX		;GET CURRENT CONTROL BITS
	  AND AL,0FFH-LPT_DIR	;MAKE SURE IT REMAINS AN OUTPUT PORT
	  OR AL,LPT_STRB	;BEGIN NEG. STROBE PULSE (OUTPUT IS INVERTED)
	  OUT DX,AL
	  OUT WAITPORT,AL	;DELAY 5us
	  OUT WAITPORT,AL
	  OUT WAITPORT,AL
	  OUT WAITPORT,AL
	  OUT WAITPORT,AL
	  AND AL,0FFH-LPT_STRB	;END STROBE PULSE
	  OUT DX,AL
	  RET

;;;
;KBINTR - PROCESS ANY PENDING KEYBOARD EVENTS, AND DISCARD ALL BUT CTRL_C.  IF
;CTRL-C WAS PRESSED, MOVE BACK IN THE EVENT SEQUENCE TO THE MOST RECENT ENDCODE
;(OR TO THE DUMMY ENDCODE IF THIS IS THE FIRST BLOCK IN THE EXPERIMENT).
;MODIFIES: AX AND ALSO REGISTERS MODIFIED BY INT 16H.  ALSO MODIFIES SI IN THE
;CASE INT WHICH THE EVENT SEQUENCE IS CHANGED.
;;;
KBINTR:	  MOV AH,01		;CHECK FOR KEY
	  INT 16H
	  JE KBINTR1
	  MOV AH,00H		;IF KEY PENDING, GET IT
	  INT 16H
	  CMP AL,03H		;IF CTRL-C, RETURN TO BEGINNING OF TRIAL
	  JNE KBINTR1
KBINTR0:  INC SI
	  MOV AL,[SI]
	  CMP AL,ENDCODE
	  JNE KBINTR0
	  INC SI
KBINTR1:  RET

;;;
;MRISYNC - PRODUCE AN ALERTING TONE UNTIL AN MRI TRIGGER PULSE OR MANUAL
;KEYBOARD INPUT IS DETECTED
;MODIFIES AL,DX, AND ALSO REGISTERS MODIFIED BY INT 16H
;;;
MRISYNC:  MOV DX,PADDR		;PARALLEL PORT BASE ADDRESS
	  INC DX		;POINT AT STATUS REGISTER
	  IN AL,PPIPORT+1	;GET CURRENT PPI SETTINGS
	  AND AL,0FFH-PPI_SENA-PPI_SDAT ;MAKE SURE SPEAKER IS OFF
	  PUSH AX		;SAVE CURRENT SETTINGS
	  OR AL,PPI_SENA+PPI_SDAT ;ENABLE SPEAKER INPUT AND TURN ON SPEAKER
	  OUT PPIPORT+1,AL
MRIWAIT:  IN AL,DX		;TEST FOR SYNC PULSE ON BUSY INPUT
	  TEST AL,LPT_BUSY
	  JNE MRIDONE
	  MOV AH,01H		;CHECK FOR KEY
	  INT 16H
	  JE MRIWAIT		;LOOP IF NO SYNC AND NO KEY
	  MOV AH,00H
	  INT 16H		;DISCARD KEY
MRIDONE:  POP AX		;RESTORE PPI SETTINGS
	  OUT PPIPORT+1,AL
	  RET

;;;
;FOPEN - OPEN A FILE FOR READING
;ON ENTRY, DX POINTS TO THE NAME OF THE FILE TO OPEN,
;AND BX POINTS TO A DATA AREA OF SIZE BUFSIZ+8.
;ON EXIT, THE FIRST FOUR WORDS OF THIS DATA AREA STORE THE FILE HANDLE,
;THE BASE ADDRESS OF THE BUFFER,
;A CURRENT-POSITION-MINUS-ONE POINTER,
;AND AN END-POSITION-PLUS-ONE POINTER, RESPECTIVELY.
;MODIFIES AX, AND ALSO REGISTERS MODIFIED BY INT 21H
;;;
FOPEN:	  MOV AX,3D00H		;OPEN READ-ONLY
	  PUSH BX		;SAVE ADDRESS OF DATA AREA
	  INT 21H
	  POP BX		;RESTORE ADDRESS OF DATA AREA
	  JC FOPENERR		;ABORT IF ERROR
	  MOV [BX],AX		;SAVE FILE HANDLE
	  MOV AX,BX
	  ADD AX,8
	  MOV [BX+2],AX		;SAVE BASE ADDRESS OF BUFFER
	  ADD AX,BUFSIZ-1	;POINT TO THE END OF THE BUFFER
	  MOV [BX+4],AX
	  INC AX		;POINT JUST PAST THE BUFFER
	  MOV [BX+6],AX
	  RET
FOPENERR: MOV AH,4CH		;TERMINATE (WITH ERROR CODE FROM OPEN IN AL)
	  INT 21H

;;;
;FREAD - BUFFERED READ OF ONE BYTE INTO AL,
;FROM THE DATA AREA POINTED TO BY BX,
;WHICH HAS PREVIOUSLY BEEN INITIALISED BY FOPEN.
;ON ATTEMPT TO READ PAST EOF, CF IS SET.  OTHERWISE CF IS CLEARED.
;MODIFIES AX,CX,DX,DI, AND ALSO REGISTERS MODIFIED BY INT 21H
;;;
FREAD:	  MOV DI,[BX+4]		;FETCH DATA POINTER
	  INC DI
	  CMP DI,[BX+6]		;SEE IF END OF BUFFER
	  JNE FREADBUF
	  SUB DI,BUFSIZ
	  CMP DI,[BX+2]		;IF ONLY A PARTIAL BUFFER WAS READ...
	  JNE FCLOSE		;...THEN WE'RE AT EOF
	  PUSH BX,DI		;SAVE POINTERS TO DATA AREA AND BUFFER
	  MOV BX,[BX]		;FILE HANDLE
	  MOV CX,BUFSIZ		;RECORD LENGTH
	  MOV DX,DI		;BUFFER
	  MOV AH,3FH		;READ
	  INT 21H
	  POP DI,BX		;RESTORE POINTERS TO DATA AREA AND BUFFER
	  JC FCLOSE		;EOF IF ERROR...
	  AND AX,AX		;...OR IF ZERO BYTES READ
	  JE FCLOSE
	  ADD AX,DI		;POINT PAST END OF BUFFER
	  MOV [BX+6],AX
FREADBUF: MOV AL,[DI]		;LOAD BYTE FROM BUFFER
	  MOV AH,0		;SUCCESS CODE
	  MOV [BX+4],DI		;SAVE NEW CURRENT-POSITION POINTER
	  CLC			;CLEAR CARRY FOR SUCCESS
	  RET
FCLOSE:	  MOV AH,3EH		;CLOSE FILE
	  MOV BX,[BX]		;FILE HANDLE
	  INT 21H
	  STC			;SET CARRY TO SIGNAL EOF
	  RET

;;;
;ISDIGIT - CLEAR CF IF AL CONTAINS AN ASCII DIGIT; OTHERWISE SET CF
;MODIFIES CF
;;;
ISDIGIT:  CMP AL,'0'		;CAN'T BE LESS THAN '0'
	  JC NOTDIGIT
	  CMP AL,'9'+1		;CAN'T BE GREATER THAN '9'
	  CMC
NOTDIGIT: RET

;;;
;SCANINT - SCAN AN 8-BIT UNSIGNED INTEGER FROM THE CURRENTLY OPEN FILE
;EXIT CONDITIONS: DL CONTAINS ACCUMULATED NUMBER,
;AH IS FFH IF EOF, 0 OTHERWISE; CF CLEAR IF DL VALID, SET OTHERWISE.
;MODIFIES AX,CX,DX,DI, AND ALSO REGISTERS MODIFIED BY INT 21H
;;;
SCANINT:  CALL FREAD		;NEXT CHARACTER INTO AL
	  JC SCANEOF0		;ABORT IF EOF BEFORE ANY DIGITS READ
	  CALL ISDIGIT
	  JC SCANINT		;SKIP WHITESPACE AND NON-DIGITS
	  MOV DL,0		;INITIALISE ACCUMULATOR TO ZERO
SCANLOOP: SUB AL,'0'		;CONVERT DIGIT TO NUMBER
	  XCHG DL,AL		;SAVE NUMBER IN DL
	  MOV DH,10
	  MUL DH		;DECIMAL SHIFT LEFT OF ACCUMULATOR (AL)
	  ADD DL,AL		;ADD SAVED NUMBER INTO ACCUMULATOR
	  CALL FREAD		;READ NEXT CHARACTER
	  JC SCANEOF1
	  CALL ISDIGIT		;IF IT'S A DIGIT, LOOP BACK AND INCLUDE IT
	  JNC SCANLOOP
	  CLC			;CLEAR CARRY TO INDICATE DL VALID
	  MOV AH,0		;CLEAR EOF FLAG
	  RET
SCANEOF1: CLC			;CLEAR CARRY TO INDICATE DL VALID
SCANEOF0: MOV AH,0FFH		;SET EOF FLAG
	  RET

;;;
;SYNC - SYNC TO VIDEO REFRESH PULSE
;MODIFIES AL,DX
;;;
SYNC:	  MOV DX,SYNCREG	;COLOUR SYNC REGISTER
SYNC1:	  IN AL,DX
	  AND AL,8H		;VSYNC
	  JNE SYNC1		;LOOP TILL VSYNC LOW
SYNC2:	  IN AL,DX
	  AND AL,8H
	  JE SYNC2		;LOOP TILL VSYNC RISES
	  RET

;;;
;JOYSTK - READ JOYSTICK VALUE INTO BX.  SET SF IF NEGATIVE, ZF IF ZERO
;MODIFIES AL,BX,DX
;;;
JOYSTK:	  MOV DX,GAMEPORT
	  MOV BX,0		;ZERO COUNTER
	  OUT DX,AL		;TRIGGER RC TIMER
JOYWAIT:  INC BX		;INCREMENT COUNTER
	  JE JOYOFLW		;STOP IF COUNTER OVERFLOWED
	  IN AL,DX		;GET STATUS
	  TEST AL,HMASK		;MASK FOR HORIZONTAL BIT
	  JNE JOYWAIT		;LOOP TILL RC TIMER IS CHARGED
JOYOFLW:  RET

;;;
;DELAY - WAIT FOUR REFRESH CYCLES, WHILE CHECKING FOR CHANGES IN JOYSTICK STATE
;MODIFIES AX,BX,DX
;NOTE: WE WOULD LIKE TO BE ABLE TO CHECK THE JOYSTICK INPUT ON EVERY SYNC,
;      BUT UNFORTUNATELY IT TAKES THE JOYSTICK CAPACITOR MORE TIME TO RECHARGE.
;;;
DELAY:	  CALL SYNC		;COUNT 2 REFRESH CYCLES
	  CALL SYNC
DELAYL:	  CALL JOYSTK		;GET JOYSTICK VALUE
	  SUB BX,[JOYOFF]	;SUBTRACT OFFSET
	  AND BL,7FH
	  MOV JOYVAL,BL
DELAY2:	  CALL SYNC		;COUNT 2 MORE REFRESH CYCLES
	  JMP SYNC

;;;
;CALIB - 2D CALIBRATION FOR ISCAN
;REPEAT CALIBRATIONS AT CENTRE & 4 CORNERS TILL SPACE BAR PRESSED FOR ALL POINTS
;MODIFIES AX,BX,CX AND ALSO REGISTERS MODIFIED BY INT 16H
;;;
;CALPAINT: PLOT A 4x4 SQUARE OF COLOUR [AH] AT OFFSET [BX]
CALPAINT: MOV CH,4
CALPNT1:  MOV CL,4
CALPNT2:  MOV [BX],AH
	  INC BX
	  DEC CL
	  JNE CALPNT2
	  ADD BX,NUMXPIX-4
	  DEC CH
	  JNE CALPNT1
	  RET
;CALPOINT: SET AND CLEAR A CALIBRATION POINT AT OFFSET [BX]
CALPOINT: MOV AH,CAL			;FOREGROUND COLOUR
	  PUSH BX
	  CALL CALPAINT			;SET THE POINT
	  MOV AH,00H
	  INT 16H			;WAIT FOR KEYBOARD (AH=00H)
	  POP BX
	  MOV AH,BGND			;BACKGROUND COLOUR
	  CALL CALPAINT			;CLEAR THE POINT
	  CMP AL,20H			;SPACEBAR INDICATES CALIBRATION OK
	  RET
CALIB:	  PUSH DS			;SAVE SEGMENT POINTER
	  MOV DS,0A000H			;VIDEO SEGMENT
	  CALL CLS			;CLEAR DISPLAY
CALIB1:	  MOV BX,40158	;CENTRE: NUMXPIX*(NUMYPIX/4+3*NUMYPIX/8) + NUMXPIX/2-2
	  CALL CALPOINT
	  JNE CALIB1
	  MOV BX,16048	;UPPER LEFT: NUMXPIX*(NUMYPIX/4) + NUMYPIX/4-2
	  CALL CALPOINT
	  JNE CALIB1
	  MOV BX,16268	;UPPER RIGHT: NUMXPIX*(NUMYPIX/4) +(NUMXPIX-NUMYPIX/4-2)
	  CALL CALPOINT
	  JNE CALIB1
	  MOV BX,62768	;LOWER LEFT: NUMXPIX*(NUMYPIX-4) + NUMYPIX/4-2
	  CALL CALPOINT
	  JNE CALIB1
	  MOV BX,62988	;LOWER RIGHT: NUMXPIX*(NUMYPIX-4) +(NUMXPIX-NUMYPIX/4-2)
	  CALL CALPOINT
	  JNE CALIB1
	  POP DS			;RESTORE SEGMENT POINTER
	  RET

;;;
;PLOT A ROW OF PIXELS OF LENGTH [BX] AT POSITION [CX,AX], OF COLOUR [DL].
;(NO BOUNDS-CHECKING IS DONE.)
;MODIFIES AX,BX,DH,DI
;;;
PLOTROW:  PUSH DS		;SAVE DATA SEGMENT BASE ADDRESS
	  MOV DS,0A000H		;POINT TO VIDEO MEMORY SEGMENT
	  PUSH DX		;SAVE COLOUR
	  MOV DX,NUMXPIX
	  MUL DX		;PLACE ROW OFFSET IN AX (HIGH-ORDER BITS ARE 0)
	  ADD AX,CX		;ADD COLUMN OFFSET INTO AX
	  DEC AX		;OFFSET [1..length] INSTEAD OF [0..length-1]
	  MOV DI,AX		;MOVE BASE ADDR INTO DI
	  POP DX		;RESTORE COLOUR
PLOTLOOP: MOV [BX][DI],DL	;SET PIXEL
	  DEC BX
	  JNE PLOTLOOP
	  POP DS		;RESTORE DATA SEGMENT BASE ADDRESS
	  RET

;;;
;DRAW A SQUARE OF COLOUR [DL] AT HORIZONTAL OFFSET [CX]
;MODIFIES AX,BX,DH,DI
;;;
SQUARE:	  MOV BX,SQ_HLEN	;LENGTH OF ROW
	  MOV AX, NUMYPIX/2 - SQ_VLEN/2	;VERTICAL COORDINATE
SQLOOP:	  PUSH AX,BX,CX
	  CALL PLOTROW
	  POP CX,BX,AX
	  INC AX
	  CMP AX, NUMYPIX/2 + SQ_VLEN/2
	  JNE SQLOOP
	  RET

;;;
;DRAW A SQUARE OF COLOUR [DL] ON THE RIGHT SIDE OF THE SCREEN
;MODIFIES AX,BX,CX,DH,DI
;;;
RSQUARE:  MOV CX,SQ_RPOS	;RIGHT MARGIN
	  JMP SQUARE

;;;
;DRAW A SQUARE OF COLOUR [DL] ON THE LEFT SIDE OF THE SCREEN
;MODIFIES AX,BX,CX,DH,DI
;;;
LSQUARE:  MOV CX,SQ_LPOS	;LEFT MARGIN
	  JMP SQUARE

;;;
;PLOT A STIMULUS:
;	AL b0 INDICATES COLOUR (0=TARGET, 1=NON-TARGET)
;	AL b2:b1:b0 INDICATES LOCATION (001,010=LEFT, 011,100=RIGHT)
;MODIFIES AX,BX,CX,DX,DI
;;;
PLOT:	  TEST AL,01H		;SEE IF TARGET
	  JE PLOTT
	  PUSH AX
	  MOV AL,[JOYVAL]	;IF NON-TARGET, SEND JOYSTICK POSITION
	  CALL SOUT
	  POP AX
	  MOV DL,NONTARG	;NON-TARGET COLOUR
	  JMP PLOTNT
PLOTT:	  CALL SOUT		;IF TARGET, SEND EVENT CODE
	  MOV DL,TARGET		;TARGET COLOUR
PLOTNT:	  MOV CX,SQ_RPOS	;RIGHT MARGIN...
	  CMP AL,3
	  JNC PLOTR
	  MOV CX,SQ_LPOS	;...UNLESS CODE 1 OR 2
PLOTR:	  CALL SQUARE		;PLOT SQUARE
	  PUSH CX
	  CALL DELAY		;WAIT
	  POP CX
	  MOV DL,BGND		;CLEAR SQUARE
	  JMP SQUARE

;;;
;DRAW THE FIXATION CROSS
;MODIFIES AX,BX,CX,DX,DI
;;;
FIXATE:	  MOV BX,(SQ_HLEN/8)*2+1     ;PLOT HORIZONTAL PORTION
	  MOV CX,NUMXPIX/2-SQ_HLEN/8
	  MOV AX,3*NUMYPIX/4+SQ_VLEN
	  MOV DL,FIX
	  CALL PLOTROW
	  MOV BX, 320*(3*NUMYPIX/4 + 7*SQ_VLEN/8) + NUMXPIX/2
	  PUSH DS		;NOW PLOTTING VERTICAL PORTION
	  MOV DS,0A000H		;POINT TO VIDEO MEMORY SEGMENT
FIXLOOP:  MOV [BX],DL		;SET PIXEL
	  ADD BX,320		;MOVE TO NEXT ROW
	  CMP BX,63840		;320*(3*NUMYPIX/4 + 9*SQ_VLEN/8 + 1) + NUMXPIX/2
	  JNE FIXLOOP
	  POP DS		;RESTORE DATA SEGMENT POINTER
	  RET

;;;
;SET ALL OF VIDEO MEMORY TO THE COLOUR 'BGND'
;MODIFIES BX
;;;
CLS:	  MOV BX,NUMXPIX*NUMYPIX-1
	  PUSH DS
	  MOV DS,0A000H
CLSLOOP:  MOV BYTE PTR [BX],BGND
	  DEC BX
	  JNE CLSLOOP
	  MOV BYTE PTR [BX],BGND
	  POP DS
	  RET

;;;
;DEBUGGING ROUTINES
;;;

;;;
;PRNTNUM - PRINT THE BASE-TEN REPRESENTATION OF [AX]
;MODIFIES AX,BX,CX,DX, AND ALSO REGISTERS MODIFIED BY INT 21H
;;;
;PRNTNUM: MOV CX,10000		;INITIAL DIVISOR IS 10^4
;PRNTDIG: MOV DX,0		;HIGH BITS OF DIVIDEND ARE ZERO
;	  DIV CX		;QUOTIENT IN AL, ZERO IN AH, REMAINDER IN DX
;	  PUSH DX		;SAVE REMAINDER
;	  PUSH CX		;SAVE DIVISOR
;	  ADD AL,'0'		;MAKE THE QUOTIENT INTO A DIGIT
;	  DEC SP		;ALLOCATE A CHARACTER BUFFER FROM THE STACK
;	  MOV BX,SP		;POINT TO IT
;	  MOV [BX],AL		;PUT DIGIT INTO BUFFER
;	  MOV AH,40H		;OUTPUT IT
;	  MOV DX,BX
;	  MOV BX,1		;TO FILE HANDLE 1 (STDOUT)
;	  MOV CX,1		;AS A SINGLE CHARACTER
;	  INT 21H
;	  INC SP		;FREE BUFFER
;	  POP AX		;RESTORE DIVISOR INTO AX
;	  MOV DX,0
;	  MOV CX,10		;DIVIDE IT BY TEN
;	  DIV CX		;NEW DIVISOR IN AX, 0 (OR 1, ON LAST LOOP) IN DX
;	  MOV CX,AX		;MOVE DIVISOR BACK INTO CX
;	  POP AX		;RESTORE REMAINDER INTO AX
;	  AND CX,CX
;	  JNE PRNTDIG		;LOOP TILL DIVISOR IS ZERO
;	  RET

;;;
;DISPNUM - DISPLAY A 16-BIT NUMBER FROM AX IN VIDEO RAM
;MODIFIES NONE
;;;
;DISPNUM: PUSHF			;SAVE FLAGS
;	  PUSHA			;SAVE GENERAL REGISTERS
;	  PUSH DS		;SAVE SEGMENT REGISTER
;	  MOV DS,0A000H		;POINT TO VIDEO SEGMENT
;	  MOV BX,NUMXPIX	;OFFSET ONE ROW
;	  MOV DI,BX
;	  MOV BX,001EH		;USE EVEN ADDRESSES 0 THROUGH 30
;DISPLOOP: ROR AX,1		;LSb INTO CARRY
;	  JNC DISPCLR
;	  MOV BYTE PTR [BX][DI],0 ;IF LSb SET, CLEAR LOWER ROW AND SET UPPER ROW
;	  MOV BYTE PTR [BX],15	;(OTHERWISE SET LOWER AND CLEAR UPPER, BELOW)
;DISPDEC: DEC BL		;LOOP TILL THROUGH WITH ALL 16 BITS
;	  MOV BYTE PTR [BX][DI], 0 ;CLEAR ODD ADDRESSES
;	  MOV BYTE PTR [BX],0
;	  DEC BL
;	  JNS DISPLOOP
;	  MOV AH,00H		;DELAY TILL KEY PRESSED
;	  INT 16H
;	  POP DS		;RESTORE SEGMENT REGISTER
;	  POPA			;RESTORE GENERAL REGISTERS
;	  POPF			;RESTORE FLAGS
;	  RET
;DISPCLR: MOV BYTE PTR [BX][DI],15
;	  MOV BYTE PTR [BX],0
;	  JMP DISPDEC

;;;
;STATIC DATA
;;;
SADDR	  DW ?			;BASE ADDRESS OF SERIAL PORT
PADDR	  DW ?			;BASE ADDRESS OF PARALLEL PORT
JOYOFF	  DW 0040H		;MIDPOINT OF JOYSTICK EVENT CODE RANGE
JOYVAL	  DB 0			;MSB OF CURRENT JOYSTICK VALUE

;;;
;END OF PROGRAM
;;;
	  ENDP
CODESG	  ENDS
	  END  MAIN
